/*
 * Program Name
 *  Msg2File - Message to a File
 *
 * Last date of modification
 *  Feb 15, 2010
 *
 * Description
 *  This program will get a message and write it to a file.
 *
 *
 *  The following is the logic in this program:
 *
 *  - Initialize and allocated memory.
 *  - Get the command line parameters.
 *  - Connect to the queue manager.
 *  - Open output queue
 *  - Loop
 *  - - Get message
 *  - - Write to a file
 *  - Close output queue
 *  - Disconnect from the queue manager.
 *  - Terminate
 *
 * Copyright(C), 2001-2010, Roger Lacroix
 * Apache 2 License
 */

#ifdef MVS

#define  _XOPEN_SOURCE_EXTENDED 1

/*
 * MVS (OS/390) only pragma(s).
 */
#pragma runopts(ENV(MVS),PLIST(MVS),NOEXECOPS,\
NOTEST(NONE,INSPICD,PROMPT) )

#pragma csect(code,"M2F_MAIN")
#pragma csect(static,"M2F_WS")
#pragma comment( copyright , "Roger Lacroix." )
#endif

#pragma subtitle( "Load module eye-catchers" )
/*
 * Program build time eye-catchers for executable.
 */
static const char * EYECATCHER_EyeCatcher  = "Msg2File " ;
static const char * EYECATCHER_Version     = "1.2.1" ;
static const char * EYECATCHER_Datestamp   = __DATE__ ;
static const char * EYECATCHER_Timestamp   = __TIME__ ;
static const char * EYECATCHER_Filename    = __FILE__ ;

/*
 * Headers
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
/* includes for MQI */
#include <cmqc.h>

#ifdef WIN32
#define  MAXPATH _MAX_PATH
#else
#define  MAXPATH PATH_MAX
#endif

/*
 * Global variables
 */
MQCHAR   QMName[MQ_Q_MGR_NAME_LENGTH];  // queue manager name
MQCHAR   QName[MQ_Q_NAME_LENGTH];       // queue manager name

char    *pBuffer;
char     GetWithConvert='N';
char     FileName[MAXPATH] = "msg2file.txt";
char     FileNameOnly[MAXPATH] = "";
char     FileExtension[MAXPATH] = "";

char* usage = " -f filename -m qmgr -q queue [-c] \n\
-f filename       The name of the output file              \n\
-m qmgr           The name of the queue manager            \n\
-q queue          The name of the output queue             \n\
-c                MQGet with convert                    [Optional]\n\
-?                Display this help message             [Optional]\n\
";

/**
 * Prototypes
 */
void GetParameters(int  argc, char *argv[]);
long FileSize (char  *);
int  WriteToFile (int, int );

/**
 * Mainline
 */
int main(int argc, char **argv)
{
   MQOD     od = {MQOD_DEFAULT};    /* Object Descriptor             */
   MQMD     md = {MQMD_DEFAULT};    /* Message Descriptor            */
   MQGMO   gmo = {MQGMO_DEFAULT};   /* get message options           */

   MQHCONN  Hcon;                   /* connection handle             */
   MQHOBJ   Hobj;                   /* object handle                 */
   MQLONG   O_options;              /* MQOPEN options                */
   MQLONG   C_options;              /* MQCLOSE options               */
   MQLONG   CompCode;               /* completion code               */
   MQLONG   OpenCode;               /* MQOPEN completion code        */
   MQLONG   Reason;                 /* reason code                   */
   MQLONG   CReason;                /* reason code for MQCONN        */
   MQLONG   buflen;                 /* buffer length                 */
   MQLONG   messlen;                /* message length received       */
   int      seqNum = 1;
   char     *extension = NULL;

   // printf( "argc=%d\n", argc );

   if ( (argc < 5) || (argc > 8) )
   {
      printf( "Usage:\n%s%s\n", argv[0], usage );
      exit(-1);
   }

   printf("Starting Msg2File\n");

   /* Clear all queue and qmgr names. */
   memset( QMName, ' ', MQ_Q_NAME_LENGTH );
   memset( QName,  ' ', MQ_Q_NAME_LENGTH );

   GetParameters(argc, argv);

   /* Connect to queue manager */
   MQCONN(QMName,                  /* queue manager                  */
          &Hcon,                   /* connection handle              */
          &CompCode,               /* completion code                */
          &CReason);               /* reason code                    */

   printf("MQCONN ended with cc %ld rc %ld\n", CompCode, CReason);
   /* report reason and stop if it failed     */
   if (CompCode == MQCC_FAILED)
   {
     exit( (int)CReason );
   }

   /* Use parameter as the name of the target queue  */
   strncpy(od.ObjectName, QName, (size_t)MQ_Q_NAME_LENGTH);

   printf("Source queue is %s\n", od.ObjectName);

   /* Open the target message queue for output */
   O_options = MQOO_INPUT_SHARED + MQOO_FAIL_IF_QUIESCING;

   MQOPEN(Hcon,                      /* connection handle            */
          &od,                       /* object descriptor for queue  */
          O_options,                 /* open options                 */
          &Hobj,                     /* object handle                */
          &OpenCode,                 /* MQOPEN completion code       */
          &Reason);                  /* reason code                  */

   /* report reason, if any; stop if failed      */
   printf("MQOPEN ended with reason code %ld\n", Reason);

   if (OpenCode == MQCC_OK)
   {
      CompCode = OpenCode;        /* use MQOPEN result for initial test */

      extension = strrchr( FileName, '.' );
      if (extension == NULL)
      {
         memcpy(FileNameOnly, FileName, sizeof(FileNameOnly));
         memset(FileExtension, '\0', sizeof(FileExtension));
      }
      else
      {
         *extension = '\0'; /* replace the period with a null */
         strncpy(FileNameOnly, FileName, sizeof(FileNameOnly));
         strncpy(FileExtension, extension+1, sizeof(FileExtension));
      }

   //   printf("FileNameOnly=%s : FileExtension=%s\n", FileNameOnly, FileExtension);

      buflen = 4 * 1024 * 1024;  // 4MB
//      buflen = 100 * 1024 * 1024;  // 100MB
      pBuffer = malloc( buflen);

      while (CompCode != MQCC_FAILED)
      {
         memset(pBuffer, '\0', buflen);  /* clear the buffer */

         memcpy(md.MsgId, MQMI_NONE, sizeof(md.MsgId));
         memcpy(md.CorrelId, MQCI_NONE, sizeof(md.CorrelId));
         md.Encoding       = MQENC_NATIVE;
         md.CodedCharSetId = MQCCSI_Q_MGR;

         gmo.Options = MQGMO_NO_WAIT + MQGMO_NO_SYNCPOINT + MQGMO_FAIL_IF_QUIESCING;

         if (GetWithConvert == 'Y')
         {
            gmo.Options += MQGMO_CONVERT;
            printf("GetWithConvert=%c\n", GetWithConvert);
         }

         MQGET(Hcon,                /* connection handle                 */
               Hobj,                /* object handle                     */
               &md,                 /* message descriptor                */
               &gmo,                /* get message options               */
               buflen,              /* pBuffer length                    */
               pBuffer,             /* pointer to message buffer         */
               &messlen,            /* message length                    */
               &CompCode,           /* completion code                   */
               &Reason);            /* reason code                       */

         /* report reason, if any     */
         if (Reason != MQRC_NONE)
         {
            if (Reason == MQRC_NO_MSG_AVAILABLE)
            {                         /* special report for normal end    */
               printf("MQGET No more messages\n");
            }
            else                      /* general report for other reasons */
            {
               printf("MQGET ended with reason code %d\n", Reason);

               /*   treat truncated message as a failure   */
               if (Reason == MQRC_TRUNCATED_MSG_FAILED)
                 CompCode = MQCC_FAILED;
            }
         }
         else
            printf("MQGET ended with reason code %d\n", Reason);

         /*
          * Save the message
          */
         if (CompCode != MQCC_FAILED)
         {
            WriteToFile(messlen, seqNum);
            seqNum++;
         }
      }

      if ( pBuffer != NULL )
      {
         free(pBuffer);
      }

      /* Close the target queue (if it was opened) */
      if (OpenCode != MQCC_FAILED)
      {
         C_options = 0;                  /* no close options             */
         MQCLOSE(Hcon,                   /* connection handle            */
                 &Hobj,                   /* object handle               */
                 C_options,
                 &CompCode,               /* completion code             */
                 &Reason);                /* reason code                 */

         printf("MQCLOSE ended with reason code %ld\n", Reason);
      }
   }
   else
   {
      printf("unable to open queue for output\n");
   }


   /* Disconnect from MQM if not already connected  */
   if (CReason != MQRC_ALREADY_CONNECTED)
   {
      MQDISC(&Hcon,                   /* connection handle            */
             &CompCode,               /* completion code              */
             &Reason);                /* reason code                  */

      printf("MQDISC ended with reason code %ld\n", Reason);
   }

   printf("Msg2File is ending.\n");
   return(0);
}

/**
 *
 */
void GetParameters(int argc, char *argv[])
{
   #define MaxParams 6

   char *ParmkeyUC[MaxParams] = { "-F", "-M", "-Q", "-C",",","-?"};
   char *ParmkeyLC[MaxParams] = { "-f", "-m", "-q", "-c",",","-?"};
   int i,j;
   int rc=0;

  /* Prior to processing the parameters, establish the default */
  /* values this is done before calling GetParameters          */

   if (argc >1)
   {
      for (i=1; i<argc; i++)
      {
         for (j=0; j<MaxParams; j++)
         {
            if ( (!strncmp( argv[i], ParmkeyUC[j], 2)) ||
                 (!strncmp( argv[i], ParmkeyLC[j], 2)) )
            {
               break; /* fall out to switch stmt */
            }
         };

         switch (j)
         {
            case 0:                      /* -f */
               if (i+1 < argc)
               {
                   strncpy( FileName, argv[i+1], MAXPATH );
                   i++;
               }
               else
               {
                   rc=9;
               }
               break;

            case 1:                      /* -m */
               if (i+1 < argc)
               {
                   strncpy(QMName, argv[i+1], MQ_Q_NAME_LENGTH);
                   i++;
               }
               else
               {
                   rc=9;
               }
               break;

            case 2:                      /* -q */
               if (i+1 < argc)
               {
                   strncpy(QName, argv[i+1], MQ_Q_NAME_LENGTH);
                   i++;
               }
               else
               {
                   rc=9;
               }
               break;

            case 3:                      /* -c */
               GetWithConvert = 'Y';
               break;

           case 4:                      /* , */
               // ignore the comma
               break;

           case 5:                      /* -? */
               printf( "Usage:\n%s%s\n", argv[0], usage );
               exit(-1);
               break;

           default:/* parameter error */
               printf( "Error with keyword parameter!\n" );
               printf( "Parameter=%s\n", argv[i] );
               exit(-1);
               break;
                /* ignore it */
         };
      };
   }

   if ( *(QName) == ' ' || *(QMName) == ' ' || *(FileName) == ' ' )
   {
      printf( "Usage:\n%s%s\n", argv[0], usage );
      exit(-1);
   }

   if ( rc==9 )
   {
      printf( "Usage:\n%s%s\n", argv[0], usage );
      exit(-1);
   }

   return;
}

/**
 * Write the message to a file.
 */
int  WriteToFile(int BuffSize, int seqNum )
{
   FILE   *stream;
   int    BytesWrttien=0;
   int    rc;
   char   *ptemp;
   char   NewFileName[MAXPATH];

   if (FileExtension[0] == '\0')
      sprintf(NewFileName,
              "%s_%.4d",
              FileNameOnly,
              seqNum);
   else
      sprintf(NewFileName,
              "%s_%.4d.%s",
              FileNameOnly,
              seqNum,
              FileExtension);

   ptemp = pBuffer;

   /* Open file to read line from: */
   if ( (stream = fopen( NewFileName, "wb+" )) == NULL )
   {
      fprintf(stderr,"ERROR - Cannot create output file (%s)\n",NewFileName);
      rc = -1;
   }
   else
   {
      while(BuffSize--)
      {
         fputc( *ptemp++, stream);
         BytesWrttien++;
      }

      fclose( stream );

      printf("Wrote %d bytes to file %s\n", BytesWrttien, NewFileName );

      rc = BytesWrttien;
   }
   return rc;
}
